{
   - test reconstruction / gros trous : \SIMDATA\DASH\DVIPERDH.FSH de NFS1
                         petits trous : \NFS2SE\GAMEDATA\RENDER\PC\FXTEXT.FSH
                                        \NFS3\GAMEDATA\DASHHUD\HUD98.FSH
                                        \NFS3\GAMEDATA\TRACKS\TRK003\SKY.FSH
}

{ SHPI.DAT file structure :
   - 'SHL1' (version tag : v1)
   - 16 bytes : SHPI header
   - 8.N bytes : SHPI directory
   - 4 bytes : glob palette name (0x0L if not present)
   - collection of (in order of FSH file) :
      * 4 bytes : bitmap tag (insignificant for a hole)
      * 4 bytes : offset in FSH (for consistency check)
      * 4 bytes : bitmap type. can be
         -1 -> hole in FSH file
          0 -> unknown : created <tag>.BIN with the whole entry
          1 -> 8 bit bitmap, followed by local palette entries, NFS3 like
          2 -> 32 bit bitmap
          3 -> 8 bit bitmap, with global palette, NFS2 like
          4 -> 15 bit bitmap, tag 0x78
          5 -> 15 bit bitmap, tag 0x7E
          6 -> 24 bit bitmap
          others to be added
      * type specific entry just after each relevant entry :
       if type -1: 4 bytes size, and data
       if type 1 : 16 byte header, 4 bytes palblock size,
                   and backup of whole palblock incl. its header
       if type 2 : 16 byte header, and alpha byte for each pixel (in BMP order !!!)
       if type 3 : 16 byte header
       if type 4/5 : 16 byte header, and alpha bit for each pixel (in BMP order, byte aligned !!!)
       if type 6 : 16 byte header

 FSH tags : $22 = 24bit DOS palette
            $24 = 24bit palette
            $2A = 32bit palette
            $2D = 15bit palette

            $7C = (not local palette !) ??? anim. data (2 dwords/item) ???
            $6F = text attachment

            $78 = 15bit BMP (NFS2)
            $7B = 8bit BMP
            $7D = 32bit BMP (NFS3)
            $7E = 15bit BMP (NFS3)
            $7F = 24bit BMP (NFS2SE)

 !!! holes are ignored if <=3 bytes, and also in 8-bit b/w bmp & loc-palette
}

uses dos;
const TYPE_UNKNOWN=0;
      TYPE_8BMP_LOCPAL=1;
      TYPE_32BMP=2;
      TYPE_8BMP_GLOBPAL=3;
      TYPE_15BMP_78=4;
      TYPE_15BMP_7E=5;
      TYPE_24BMP=6;
      TYPE_HOLE=-1;

type mypal=array[0..255] of longint;

var fsh:file;
    bmp:file;
    log:file;
    pal15:array[0..255] of word;
    pal24,globpal24,bmppal:mypal;
    i,x,j,code,enlen,nbmp,refofs,nxofs,width,height,totsz,pallen,bmpw:longint;
    curofs,globpallen:longint;
    c:word;
    globpalname:string[5];
    isglobpal:boolean;
    fname,s:string;
    buffer:array[0..16383] of byte;
    buf15:array[0..8191] of word absolute buffer;
    bmphead:record
       headb,headm:char;
       size:longint;
       null,ofsbmp:longint;
       hsz,wid,hei:longint;
       planes,bpp:word;
       compr,imsz,xpel,ypel,clrs,clri:longint;
    end;

    shdir:array[0..1000,0..1] of longint;
    auxbuf:array[0..16383] of byte;
    bisauxbuf:array[0..8191] of byte;
    fshhead:array[0..15] of byte;
    fshhead16:array[0..7] of word absolute fshhead;

procedure abort(msg:string);
begin
     writeln;
     writeln('ERROR ! ',msg,' - Aborting.');
     close(fsh); erase(fsh);
     halt;
end;

procedure copybytes(var f1,f2:file;len:longint);
begin
     while len>16384 do begin
       blockread(f1,buffer,16384);
       blockwrite(f2,buffer,16384);
       dec(len,16384);
     end;
     if len>0 then begin
       blockread(f1,buffer,len);
       blockwrite(f2,buffer,len);
     end;
end;

function p15to24(x:word):longint;
var r,g,b:longint;
begin
     b:=(x and $1F) shl 3;
     g:=((x shr 5) and $1F) shl 3;
     r:=((x shr 10) and $1F) shl 3;
     p15to24:=r shl 16+g shl 8+b;
end;

procedure makepal(var f:file;var len:longint;var pal:mypal);
var x,code:longint;
begin
     blockread(f,code,4);
     code:=code and $FF;
     len:=0;
     blockread(f,len,2);
     if code<>$7C then blockread(f,buffer,10);
{ code $24 }
     if code=$24 then begin
       for x:=0 to len-1 do begin
         blockread(f,buffer,3);
         pal[x]:=longint(buffer[0]) shl 16+longint(buffer[1]) shl 8+
                     longint(buffer[2]);
       end;
     end else
{ code $22 }
     if code=$22 then begin
       for x:=0 to len-1 do begin
         blockread(f,buffer,3);
         pal[x]:=longint(buffer[0]) shl 18+longint(buffer[1]) shl 10+
                     longint(buffer[2]) shl 2;
       end;
     end else
{ code $2D }
     if code=$2D then begin
       blockread(f,pal15,2*len);
       for x:=0 to len-1 do pal[x]:=p15to24(pal15[x]);
     end else
{ code $2A }
     if code=$2A then begin
       blockread(f,pal,4*len);
     end else
{ code $7C and globpal }
     if (code=$7C) and isglobpal then begin
       len:=globpallen;
       for x:=0 to len-1 do pal[x]:=globpal24[x];
     end else
{ unknown would-be palette }
     begin
       writeln('Unknown palette structure !!! Filling with black');
       len:=256;
       for x:=0 to 255 do pal[x]:=0;
     end;
{ looks like Paint prefers 256 colors everytime... sorry ! }
     if len<256 then for x:=len to 255 do pal[x]:=0;
     len:=256;
end;

begin
     writeln;
     writeln('BMP2FSH - Version 1.0 - (c) Denis Auroux 1998 (auroux@clipper.ens.fr)');
     writeln('=====================================================================');
     if paramcount<2 then begin
       writeln('Usage: BMP2FSH dirname file.FSH');
       halt;
     end;
{$I-}
     assign(log,paramstr(1)+'\SHPI.DAT');
     reset(log,1);
     if ioresult<>0 then begin
       writeln('ERROR: directory does not exist or does not contain SHPI.DAT');
       halt;
     end;
     s[0]:=#8; blockread(log,s[1],8);
     if s<>'SHL1SHPI' then begin
       writeln('SHPI.DAT does not have a valid header !');
       halt;
     end;
     fname:=paramstr(2);
     x:=0;
     for i:=1 to length(fname) do begin
       fname[i]:=upcase(fname[i]);
       if fname[i]='.' then x:=i;
       if fname[i]='\' then x:=0;
     end;
     if x=0 then fname:=fname+'.FSH';
     assign(fsh,fname);
     rewrite(fsh,1);
     if ioresult<>0 then begin
       writeln('Cannot create FSH file !');
       halt;
     end;
{$I+}
     s:='SHPI';
     blockwrite(fsh,s[1],4);
     blockread(log,totsz,4);
     blockread(log,nbmp,4);
     blockwrite(fsh,totsz,4);
     blockwrite(fsh,nbmp,4);
     writeln(nbmp,' objects in structure.');
     blockread(log,i,4); { directory id }
     blockwrite(fsh,i,4);
     blockread(log,shdir,8*nbmp);
     blockwrite(fsh,shdir,8*nbmp);
     globpalname[0]:=#4;
     blockread(log,globpalname[1],4);
     isglobpal:=(globpalname[1]<>#0);
     if isglobpal then begin
       writeln('Reading the global palette from ',globpalname,'.BIN');
{$I-}
       assign(bmp,paramstr(1)+'\'+globpalname+'.BIN');
       reset(bmp,1);
       if ioresult<>0 then abort('Cannot find the global palette');
{$I+}
       makepal(bmp,globpallen,globpal24);
       close(bmp);
     end;

{ main loop }
     curofs:=16+8*nbmp;
     while curofs<totsz do begin
       if totsz-curofs<=3 then begin
         x:=0; blockwrite(fsh,x,totsz-curofs);
         close(fsh); close(log);
         halt;
       end;
       blockread(log,s[1],4); s[0]:=#4;
       blockread(log,refofs,4);
       blockread(log,code,4);
       fname:=paramstr(1)+'\'+s;
       if (refofs<curofs) or (refofs>curofs+3) then
         abort('Incoherent byte offset');
       if curofs<refofs then begin
         x:=0; blockwrite(fsh,x,refofs-curofs); { padding }
         curofs:=refofs;
       end;
       if code=TYPE_HOLE then begin { holes }
          blockread(log,x,4);
          writeln('Adding hole');
          copybytes(log,fsh,x);
          inc(curofs,x);
       end else
       if code=TYPE_UNKNOWN then begin { unknown data }
          writeln('Adding ',s,'.BIN (binary data)');
{$I-}
          assign(bmp,fname+'.BIN');
          reset(bmp,1);
          if ioresult<>0 then abort('Cannot find file');
{$I+}
          inc(curofs,filesize(bmp));
          copybytes(bmp,fsh,filesize(bmp));
          close(bmp);
       end else
       begin {bitmap}
          if code>6 then abort('Invalid type tag encountered');
          fname:=fname+'.BMP';
          writeln('Adding ',s,'.BMP');
{$I-}
          assign(bmp,fname);
          reset(bmp,1);
          if ioresult<>0 then abort('Cannot find file');
{$I+}
          blockread(bmp,bmphead,sizeof(bmphead));
          if (bmphead.headb<>'B') or (bmphead.headm<>'M') then
            abort('Not a valid BMP file');
          blockread(log,fshhead,16);
          blockwrite(fsh,fshhead,16);
          { basic sanity checks }
          if bmphead.planes<>1 then abort('Unsupported BMP format');
          width:=fshhead16[2]; height:=fshhead16[3];
          if (bmphead.wid<>width) or (bmphead.hei<>height) then
            abort('Incorrect BMP geometry (width and/or height)');
          if (bmphead.compr<>0) then abort('Compressed BMP files not allowed');
          case bmphead.bpp of
            8:case code of
                TYPE_8BMP_LOCPAL:
                  begin
                    enlen:=fshhead[1]+longint(fshhead16[1]) shl 8;
                    bmpw:=width; while (bmpw and 3)<>0 do inc(bmpw);
                    for x:=height-1 downto 0 do begin
                      seek(bmp,bmphead.ofsbmp+x*bmpw);
                      blockread(bmp,buffer,width);
                      blockwrite(fsh,buffer,width);
                    end;
                    inc(curofs,enlen);
                    dec(enlen,width*height+16);
                    if (enlen<0) or (enlen>7) then
                      abort('Abnormal padding before local palette');
                    fillchar(buffer,8,0);
                    if enlen>0 then blockwrite(fsh,buffer,enlen);
                    blockread(log,enlen,4);
                    x:=filepos(log);
                    makepal(log,pallen,pal24);
                    seek(log,x);
                    inc(curofs,enlen);
                    copybytes(log,fsh,enlen);
                    { palette sanity check }
                    seek(bmp,sizeof(bmphead));
                    blockread(bmp,bmppal,4*pallen);
                    for x:=0 to pallen-1 do
                      if (pal24[x] and $FFFFFF)<>(bmppal[x] and $FFFFFF)
                        then abort('BMP palette has been modified');
                  end;
                TYPE_8BMP_GLOBPAL:
                  begin
                    bmpw:=width; while (bmpw and 3)<>0 do inc(bmpw);
                    for x:=height-1 downto 0 do begin
                      seek(bmp,bmphead.ofsbmp+x*bmpw);
                      blockread(bmp,buffer,width);
                      blockwrite(fsh,buffer,width);
                    end;
                    inc(curofs,width*height+16);
                    { palette sanity check }
                    seek(bmp,sizeof(bmphead));
                    blockread(bmp,bmppal,4*pallen);
                    for x:=0 to pallen-1 do
                      if (globpal24[x] and $FFFFFF)<>(bmppal[x] and $FFFFFF)
                        then abort('BMP palette has been modified');
                  end;
                else abort('Incorrect BMP color depth (8)');
              end;
            24:case code of
                TYPE_15BMP_78,TYPE_15BMP_7E:
                  begin
                    bmpw:=3*width; while (bmpw and 3)<>0 do inc(bmpw);
                    enlen:=filepos(log);
                    pallen:=(width+7) shr 3;
                    for x:=height-1 downto 0 do begin
                      seek(bmp,bmphead.ofsbmp+x*bmpw);
                      seek(log,enlen+x*pallen);
                      blockread(bmp,auxbuf,3*width);
                      blockread(log,bisauxbuf,pallen);
                      for j:=0 to width-1 do
                        if code=TYPE_15BMP_7E then { xrrrrrgggggbbbbb }
                          buf15[j]:=word(auxbuf[3*j] shr 3)+
                            word(auxbuf[3*j+1] shr 3) shl 5+
                            word(auxbuf[3*j+2] shr 3) shl 10+
                            word(((bisauxbuf[j shr 3]) shl (j and 7)) and 128) shl 8
                        else                       { rrrrrgggggxbbbbb }
                          buf15[j]:=word(auxbuf[3*j] shr 3)+
                            word(auxbuf[3*j+1] shr 3) shl 6+
                            word(auxbuf[3*j+2] shr 3) shl 11+
                            word(((bisauxbuf[j shr 3]) shl (j and 7)) and 128) shr 2;
                      blockwrite(fsh,buf15,2*width);
                    end;
                    seek(log,enlen+height*pallen);
                    inc(curofs,2*width*height+16);
                  end;
                TYPE_24BMP:
                  begin
                    bmpw:=3*width; while (bmpw and 3)<>0 do inc(bmpw);
                    for x:=height-1 downto 0 do begin
                      seek(bmp,bmphead.ofsbmp+x*bmpw);
                      blockread(bmp,buffer,3*width);
                      blockwrite(fsh,buffer,3*width);
                    end;
                    inc(curofs,3*width*height+16);
                  end;
                TYPE_32BMP:
                  begin
                    bmpw:=3*width; while (bmpw and 3)<>0 do inc(bmpw);
                    enlen:=filepos(log);
                    for x:=height-1 downto 0 do begin
                      seek(bmp,bmphead.ofsbmp+x*bmpw);
                      seek(log,enlen+x*width);
                      blockread(bmp,auxbuf,3*width);
                      blockread(log,bisauxbuf,width);
                      for j:=0 to width-1 do begin
                        buffer[4*j]:=auxbuf[3*j];
                        buffer[4*j+1]:=auxbuf[3*j+1];
                        buffer[4*j+2]:=auxbuf[3*j+2];
                        buffer[4*j+3]:=bisauxbuf[j];
                      end;
                      blockwrite(fsh,buffer,4*width);
                    end;
                    seek(log,enlen+height*width);
                    inc(curofs,4*width*height+16);
                  end;
                else abort('Incorrect BMP color depth (24)');
               end;
            else abort('Unsupported BMP color depth');
          end;
          close(bmp);
       end;
     end;
end.
