{ V1.01: fixed a name uniqueness bug by changing BMP naming convention
  (now sequence numbers instead of block ID's ; should be ok with bmp2fsh)
}

{ SHPI.DAT file structure :
   - 'SHL1' (version tag : v1)
   - 16 bytes : SHPI header
   - 8.N bytes : SHPI directory
   - 4 bytes : glob palette name (0x0L if not present)
   - collection of (in order of FSH file) :
      * 4 bytes : bitmap tag (insignificant for a hole)
      * 4 bytes : offset in FSH (for consistency check)
      * 4 bytes : bitmap type. can be
         -1 -> hole in FSH file
          0 -> unknown : created <tag>.BIN with the whole entry
          1 -> 8 bit bitmap, followed by local palette entries, NFS3 like
          2 -> 32 bit bitmap
          3 -> 8 bit bitmap, with global palette, NFS2 like
          4 -> 15 bit bitmap, tag 0x78
          5 -> 15 bit bitmap, tag 0x7E
          6 -> 24 bit bitmap
          others to be added
      * type specific entry just after each relevant entry :
       if type -1: 4 bytes size, and data
       if type 1 : 16 byte header, 4 bytes palblock size,
                   and backup of whole palblock incl. its header
       if type 2 : 16 byte header, and alpha byte for each pixel (in BMP order !!!)
       if type 3 : 16 byte header
       if type 4/5 : 16 byte header, and alpha bit for each pixel (in BMP order, byte aligned !!!)
       if type 6 : 16 byte header

 FSH tags : $22 = 24bit DOS palette
            $24 = 24bit palette
            $2A = 32bit palette
            $2D = 15bit palette

            $7C = (not local palette !) ??? anim. data (2 dwords/item) ???
            $6F = text attachment

            $78 = 15bit BMP (NFS2)
            $7B = 8bit BMP
            $7D = 32bit BMP (NFS3)
            $7E = 15bit BMP (NFS3)
            $7F = 24bit BMP (NFS2SE)

 !!! holes are ignored if <=3 bytes, and also in 8-bit b/w bmp & loc-palette
}

uses dos;
const TYPE_UNKNOWN=0;
      TYPE_8BMP_LOCPAL=1;
      TYPE_32BMP=2;
      TYPE_8BMP_GLOBPAL=3;
      TYPE_15BMP_78=4;
      TYPE_15BMP_7E=5;
      TYPE_24BMP=6;
      TYPE_HOLE=-1;

type mypal=array[0..255] of longint;

var fsh:file;
    bmp:file;
    log:file;
    pal15:array[0..255] of word;
    pal24,globpal24:mypal;
    i,x,j,code,enlen,nbmp,refofs,nxofs,width,height,totsz,pallen,bmpw:longint;
    curofs,globpallen:longint;
    c:word;
    globpalname:string[5];
    isglobpal,isglobstdname:boolean;
    fname,s:string;
    buffer:array[0..16383] of byte;
    buf15:array[0..8191] of word absolute buffer;
    bmphead:record
       headb,headm:char;
       size:longint;
       null,ofsbmp:longint;
       hsz,wid,hei:longint;
       planes,bpp:word;
       compr,imsz,xpel,ypel,clrs,clri:longint;
    end;

    shdir:array[0..1000,0..1] of longint;
    auxbuf:array[0..16383] of byte;

procedure copybytes(var f1,f2:file;len:longint);
begin
     while len>16384 do begin
       blockread(f1,buffer,16384);
       blockwrite(f2,buffer,16384);
       dec(len,16384);
     end;
     if len>0 then begin
       blockread(f1,buffer,len);
       blockwrite(f2,buffer,len);
     end;
end;

function reachnext(ffs:longint):boolean;
{ sets global vars. i, refofs, s, nxofs }
var j:longint;
begin
     refofs:=totsz; i:=-1; s[0]:=#4;
     for j:=0 to nbmp-1 do
       if (shdir[j,1]>=ffs) and (shdir[j,1]<refofs) then
         begin refofs:=shdir[j,1]; i:=j; end;
{ removed in V1.01:    move(shdir[i,0],s[1],4); }
{ added in V1.01: new naming convention }
     s[1]:=char(48+(i div 1000) mod 10);
     s[2]:=char(48+(i div 100) mod 10);
     s[3]:=char(48+(i div 10) mod 10);
     s[4]:=char(48+i mod 10);

     nxofs:=totsz;
     for j:=0 to nbmp-1 do
       if (shdir[j,1]>refofs) and (shdir[j,1]<nxofs) then
         nxofs:=shdir[j,1];

     if refofs-ffs>3 then begin { write a hole block for >3 bytes padding }
       j:=0; blockwrite(log,j,4); { any }
       blockwrite(log,ffs,4);
       j:=TYPE_HOLE; blockwrite(log,j,4);
       j:=refofs-ffs; blockwrite(log,j,4);
       seek(fsh,ffs);
       copybytes(fsh,log,refofs-ffs);
       writeln('Hole (',refofs-ffs,' bytes)');
     end;
     reachnext:=(refofs<totsz);
end;

function p15to24(x:word):longint;
var r,g,b:longint;
begin
     b:=(x and $1F) shl 3;
     g:=((x shr 5) and $1F) shl 3;
     r:=((x shr 10) and $1F) shl 3;
     p15to24:=r shl 16+g shl 8+b;
end;

procedure makepal(offs:longint;var len:longint;var pal:mypal);
var x,code:longint;
begin
     seek(fsh,offs);
     blockread(fsh,code,4);
     code:=code and $FF;
     len:=0;
     blockread(fsh,len,2);
     if code<>$7C then blockread(fsh,buffer,10);
{ code $24 }
     if code=$24 then begin
       for x:=0 to len-1 do begin
         blockread(fsh,buffer,3);
         pal[x]:=longint(buffer[0]) shl 16+longint(buffer[1]) shl 8+
                     longint(buffer[2]);
       end;
     end else
{ code $22 }
     if code=$22 then begin
       for x:=0 to len-1 do begin
         blockread(fsh,buffer,3);
         pal[x]:=longint(buffer[0]) shl 18+longint(buffer[1]) shl 10+
                     longint(buffer[2]) shl 2;
       end;
     end else
{ code $2D }
     if code=$2D then begin
       blockread(fsh,pal15,2*len);
       for x:=0 to len-1 do pal[x]:=p15to24(pal15[x]);
     end else
{ code $2A }
     if code=$2A then begin
       blockread(fsh,pal,4*len);
     end else
{ code $7C and globpal }
     if (code=$7C) and isglobpal then begin
       len:=globpallen;
       for x:=0 to len-1 do pal[x]:=globpal24[x];
     end else
{ unknown would-be palette }
     begin
       writeln('Unknown palette structure !!! Filling with black');
       len:=256;
       for x:=0 to 255 do pal[x]:=0;
     end;
{ looks like Paint prefers 256 colors everytime... sorry ! }
     if len<256 then for x:=len to 255 do pal[x]:=0;
     len:=256;
end;

begin
     writeln;
     writeln('FSH2BMP - Version 1.01 - (c) Denis Auroux 1998 (auroux@clipper.ens.fr)');
     writeln('======================================================================');
     if paramcount<2 then begin
       writeln('Usage: FSH2BMP file.FSH dirname');
       halt;
     end;
     fillchar(bmphead,sizeof(bmphead),0);
{$I-}
     assign(fsh,paramstr(1));
     reset(fsh,1);
     if ioresult<>0 then begin
       assign(fsh,paramstr(1)+'.FSH');
       reset(fsh,1);
       if ioresult<>0 then begin
         writeln('FSH file not found !');
         halt;
       end;
     end;
     s[0]:=#4; blockread(fsh,s[1],4);
     if s<>'SHPI' then begin
       writeln('This is not a FSH file !');
       halt;
     end;
     s:=paramstr(2);
     if s[length(s)]='\' then dec(s[0]);
     rmdir(s);
     i:=ioresult;
     mkdir(s);
     if ioresult<>0 then begin
       writeln('ERROR: target directory is not empty, or could not be created.');
       writeln('For safety reasons, conversion is only possible to an empty directory.');
       halt;
     end;
{$I+}
     writeln('Creating SHPI.DAT (FSH structure information file)');
     fname:=paramstr(2)+'\SHPI.DAT';
     assign(log,fname);
     rewrite(log,1);
     fname:='SHL1'; { identifier for v1 }
     blockwrite(log,fname[1],4);
     s:='SHPI';
     blockwrite(log,s[1],4);
     blockread(fsh,totsz,4);
     blockread(fsh,nbmp,4);
     blockwrite(log,totsz,4);
     blockwrite(log,nbmp,4);
     writeln(nbmp,' objects in structure.');
     blockread(fsh,i,4); { directory id }
     blockwrite(log,i,4);
     blockread(fsh,shdir,8*nbmp);
     blockwrite(log,shdir,8*nbmp);
     { first look for a global palette ; priority is name !pal }
     isglobpal:=false; isglobstdname:=false;
     for i:=0 to nbmp-1 do if not(isglobstdname) then begin
       s[0]:=#4; move(shdir[i,0],s[1],4);
       refofs:=shdir[i,1];
       seek(fsh,refofs);
       blockread(fsh,code,4);
       code:=code and $FF;
       if ((code=$24) or (code=$22) or (code=$2D) or (code=$2A)) then begin
         if s='!pal' then isglobstdname:=true;
         globpalname:=s;
         isglobpal:=true;
         makepal(refofs,globpallen,globpal24);
       end;
     end;
     if isglobpal then begin
       writeln('directory contains a global palette (',globpalname,')');
       blockwrite(log,globpalname[1],4);
     end else begin
       i:=0; blockwrite(log,i,4);
     end;

{ main loop }
     curofs:=16+8*nbmp;
     while reachnext(curofs) do begin
       blockwrite(log,s[1],4);
       blockwrite(log,refofs,4);
       write(s,' : ');
       seek(fsh,refofs);
       blockread(fsh,code,4);
       enlen:=code shr 8;
       code:=code and $FF;
       pallen:=nxofs-refofs-enlen;
       curofs:=nxofs;
       { enlen is actually the offset to the next nice place !!! }
       { lines are apparently not aligned, only after end of bmp/pal ! }

{ NFS3 8-bit bitmaps }
       if ((code=$7B) and (enlen>0) and (pallen>0)) then begin
         width:=0; height:=0;
         blockread(fsh,width,2);
         blockread(fsh,height,2);
         x:=TYPE_8BMP_LOCPAL;
         blockwrite(log,x,4);
         seek(fsh,refofs);
         blockread(fsh,buffer,16);
         blockwrite(log,buffer,16);
         blockwrite(log,pallen,4);
         seek(fsh,refofs+enlen);
         copybytes(fsh,log,pallen);
         makepal(refofs+enlen,pallen,pal24);
         writeln('8-bit bitmap with local palette (',width,'x',height,')');
         fname:=paramstr(2);
         fname:=fname+'\'+s+'.BMP';
         assign(bmp,fname);
         rewrite(bmp,1);
         bmpw:=width;
         while (bmpw and 3)<>0 do inc(bmpw);
         { writing a bitmap header }
         with bmphead do begin
           headb:='B'; headm:='M';
           size:=sizeof(bmphead)+4*pallen+bmpw*height;
           ofsbmp:=sizeof(bmphead)+4*pallen;
           hsz:=40; { header size }
           wid:=width;
           hei:=height;
           planes:=1;
           bpp:=8;
           imsz:=bmpw*height;
           clrs:=0;
           clri:=0;
           compr:=0;
         end;
         blockwrite(bmp,bmphead,sizeof(bmphead));
         blockwrite(bmp,pal24,4*pallen);
         for x:=height-1 downto 0 do begin
           {assume nothing needs to be aligned ; might need width+1 ?}
           seek(fsh,refofs+16+x*width);
           blockread(fsh,buffer,width);
           blockwrite(bmp,buffer,bmpw);
         end;
         close(bmp);
       end else

{ NFS2 8-bit bitmaps }
       if ((code=$7B) and ((enlen=0) or (pallen=0)) and isglobpal) then begin
         width:=0; height:=0;
         blockread(fsh,width,2);
         blockread(fsh,height,2);
         x:=TYPE_8BMP_GLOBPAL;
         blockwrite(log,x,4);
         seek(fsh,refofs);
         blockread(fsh,buffer,16);
         blockwrite(log,buffer,16);
         writeln('8-bit bitmap with global palette (',width,'x',height,')');
         fname:=paramstr(2);
         fname:=fname+'\'+s+'.BMP';
         assign(bmp,fname);
         rewrite(bmp,1);
         bmpw:=width;
         while (bmpw and 3)<>0 do inc(bmpw);
         { writing a bitmap header }
         with bmphead do begin
           headb:='B'; headm:='M';
           size:=sizeof(bmphead)+4*globpallen+bmpw*height;
           ofsbmp:=sizeof(bmphead)+4*globpallen;
           hsz:=40; { header size }
           wid:=width;
           hei:=height;
           planes:=1;
           bpp:=8;
           imsz:=bmpw*height;
           clrs:=0;
           clri:=0;
           compr:=0;
         end;
         blockwrite(bmp,bmphead,sizeof(bmphead));
         blockwrite(bmp,globpal24,4*globpallen);
         for x:=height-1 downto 0 do begin
           {assume nothing needs to be aligned ; might need width+1 ?}
           seek(fsh,refofs+16+x*width);
           blockread(fsh,buffer,width);
           blockwrite(bmp,buffer,bmpw);
         end;
         close(bmp);
         curofs:=refofs+16+width*height;
       end else

{ NFS3 32-bit bitmaps }
       if (code=$7D) then begin
         width:=0; height:=0;
         blockread(fsh,width,2);
         blockread(fsh,height,2);
         x:=TYPE_32BMP;
         blockwrite(log,x,4);
         seek(fsh,refofs);
         blockread(fsh,buffer,16);
         blockwrite(log,buffer,16);
         writeln('32-bit bitmap (',width,'x',height,')');
         fname:=paramstr(2);
         fname:=fname+'\'+s+'.BMP';
         assign(bmp,fname);
         rewrite(bmp,1);
         bmpw:=3*width;
         while (bmpw and 3)<>0 do inc(bmpw); { !!! here bmpw is not a width }
         { writing a bitmap header }
         with bmphead do begin
           headb:='B'; headm:='M';
           size:=sizeof(bmphead)+bmpw*height;
           ofsbmp:=sizeof(bmphead);
           hsz:=40; { header size }
           wid:=width;
           hei:=height;
           planes:=1;
           bpp:=24;
           imsz:=bmpw*height;
           clrs:=0;
           clri:=0;
           compr:=0;
         end;
         blockwrite(bmp,bmphead,sizeof(bmphead));
         for x:=height-1 downto 0 do begin
           seek(fsh,refofs+16+4*x*width);
           blockread(fsh,buffer,4*width);
           for j:=0 to width-1 do begin
             buffer[3*j]:=buffer[4*j];
             buffer[3*j+1]:=buffer[4*j+1];
             buffer[3*j+2]:=buffer[4*j+2];
             auxbuf[j]:=buffer[4*j+3];
           end;
           blockwrite(bmp,buffer,bmpw);
           blockwrite(log,auxbuf,width);
         end;
         close(bmp);
         curofs:=refofs+16+4*width*height;
       end else

{ NFS2SE 24-bit bitmaps }
       if (code=$7F) then begin
         width:=0; height:=0;
         blockread(fsh,width,2);
         blockread(fsh,height,2);
         x:=TYPE_24BMP;
         blockwrite(log,x,4);
         seek(fsh,refofs);
         blockread(fsh,buffer,16);
         blockwrite(log,buffer,16);
         writeln('24-bit bitmap (',width,'x',height,')');
         fname:=paramstr(2);
         fname:=fname+'\'+s+'.BMP';
         assign(bmp,fname);
         rewrite(bmp,1);
         bmpw:=3*width;
         while (bmpw and 3)<>0 do inc(bmpw); { !!! here bmpw is not a width }
         { writing a bitmap header }
         with bmphead do begin
           headb:='B'; headm:='M';
           size:=sizeof(bmphead)+bmpw*height;
           ofsbmp:=sizeof(bmphead);
           hsz:=40; { header size }
           wid:=width;
           hei:=height;
           planes:=1;
           bpp:=24;
           imsz:=bmpw*height;
           clrs:=0;
           clri:=0;
           compr:=0;
         end;
         blockwrite(bmp,bmphead,sizeof(bmphead));
         for x:=height-1 downto 0 do begin
           seek(fsh,refofs+16+3*x*width);
           blockread(fsh,buf15,3*width);
           blockwrite(bmp,buf15,bmpw);
         end;
         close(bmp);
         curofs:=refofs+16+3*width*height;
       end else

{ NFS2/NFS3 15-bit bitmaps }
       if ((code=$78) or (code=$7E)) then begin
         width:=0; height:=0;
         blockread(fsh,width,2);
         blockread(fsh,height,2);
         if code=$78 then x:=TYPE_15BMP_78 else x:=TYPE_15BMP_7E;
         blockwrite(log,x,4);
         seek(fsh,refofs);
         blockread(fsh,buffer,16);
         blockwrite(log,buffer,16);
         if code=$78 then
           writeln('16-bit bitmap with specific coding (',width,'x',height,')')
         else
           writeln('16-bit bitmap with Win32 coding (',width,'x',height,')');
         fname:=paramstr(2);
         fname:=fname+'\'+s+'.BMP';
         assign(bmp,fname);
         rewrite(bmp,1);
         bmpw:=3*width;
         while (bmpw and 3)<>0 do inc(bmpw); { !!! here bmpw is not a width }
         { writing a bitmap header }
         with bmphead do begin
           headb:='B'; headm:='M';
           size:=sizeof(bmphead)+bmpw*height;
           ofsbmp:=sizeof(bmphead);
           hsz:=40; { header size }
           wid:=width;
           hei:=height;
           planes:=1;
           bpp:=24;
           imsz:=bmpw*height;
           clrs:=0;
           clri:=0;
           compr:=0;
         end;
         blockwrite(bmp,bmphead,sizeof(bmphead));
         for x:=height-1 downto 0 do begin
           seek(fsh,refofs+16+2*x*width);
           blockread(fsh,buf15,2*width);
           if code=$78 then
             for j:=0 to width-1 do { rrrrrgggggxbbbbb -> xrrrrrgggggbbbbb }
               buf15[j]:=(buf15[j] and $FFC0) shr 1
                   +(buf15[j] and $1F) + (buf15[j] and $20) shl 10;
           for j:=0 to width-1 do begin
             auxbuf[3*j]:=(buf15[j] and $1F) shl 3;
             auxbuf[3*j+1]:=((buf15[j] shr 5) and $1F) shl 3;
             auxbuf[3*j+2]:=((buf15[j] shr 10) and $1F) shl 3;
           end;
           blockwrite(bmp,auxbuf,bmpw);
           for j:=0 to width-1 do begin
             if (j and 7)=0 then auxbuf[j shr 3]:=0;
             if (buf15[j] and $8000)<>0 then
                inc(auxbuf[j shr 3],128 shr (j and 7));
           end;
           blockwrite(log,auxbuf,(width+7) shr 3);
         end;
         close(bmp);
         curofs:=refofs+16+2*width*height;
       end else

{ Not one of the above bitmaps }
       begin
         x:=TYPE_UNKNOWN;
         blockwrite(log,x,4);
         if code=$24 then writeln('24-bit global palette.')
         else if code=$22 then writeln('24-bit DOS global palette.')
         else if code=$2D then writeln('15-bit global palette.')
         else if code=$2A then writeln('32-bit global palette.')
         else if code=$7B then writeln('8-bit bitmap with no palette.')
         else writeln('Unknown data type (code ',code,')');
         fname:=paramstr(2);
         fname:=fname+'\'+s+'.BIN';
         assign(bmp,fname);
         rewrite(bmp,1);
         seek(fsh,refofs);
         copybytes(fsh,bmp,nxofs-refofs);
         close(bmp);
       end;
     end;
     close(log);
end.